#include <iostream>
#include <cstdlib>

#define N_BLOCKS			32  // number of blocks
#define N_THREADS_PER_BLOCK	32  // number of threads per block
#define N_RN				(50 * N_BLOCKS * N_THREADS_PER_BLOCK)

#define ITERATIONS			10000 // number of runs

__device__ unsigned Tausworthe88(unsigned &z1, unsigned &z2, unsigned &z3)
{
	// Three-step generator with period 2^88
	unsigned b = (((z1 << 13) ^ z1) >> 19);
	z1 = (((z1 & 4294967294) << 12) ^ b);

	b = (((z2 << 2) ^ z2) >> 25);
	z2 = (((z2 & 4294967288) << 4)  ^ b);

	b = (((z3 << 3) ^ z3) >> 11);
	z3 = (((z3 & 4294967280) << 17) ^ b);

	return z1 ^ z2 ^ z3;
}

__device__ unsigned LCRNG(unsigned &z)  
{  
	const unsigned a = 1664525, c = 1013904223;
	return z = a * z + c;
}

__device__ float TauswortheLCRNG(unsigned &z1, unsigned &z2, unsigned &z3, unsigned &z)
{
	// combine both generators and normalize 0...2^32 to 0...1
	return (Tausworthe88(z1, z2, z3) ^ LCRNG(z)) * 2.3283064365e-10;
}

// Kernel:
__global__ void createRandomNumbers(float* seeds, float* randoms)
{
	int x  = blockIdx.x * blockDim.x + threadIdx.x;

	// get seed values; 2^32-1 = 4294967295; using CUDA type casting to unsigned int
	unsigned z1 = __float2uint_rn(4294967295 * seeds[4*x    ]);	// Tausworthe seeds
	unsigned z2 = __float2uint_rn(4294967295 * seeds[4*x + 1]);
	unsigned z3 = __float2uint_rn(4294967295 * seeds[4*x + 2]);
	unsigned z  = __float2uint_rn(4294967295 * seeds[4*x + 3]);	// LCRNG seed

	// number of RNs to produce in this thread:
	const unsigned n  = N_RN / N_BLOCKS / N_THREADS_PER_BLOCK;

	// create RNs and immediately copy them to global memory:
	for(unsigned i = 0; i<n; i++)
		randoms[n*x + i] = TauswortheLCRNG(z1, z2, z3, z);
}


// Host function (CPU Code)
int main (int argc, char const* argv[])
{
	// create 4 seed values for each thread:
	float seeds[4 * N_BLOCKS * N_THREADS_PER_BLOCK];
	for(unsigned i=0; i<(4 * N_BLOCKS * N_THREADS_PER_BLOCK); i++)
		 seeds[i] = drand48();

	// create array that will finally get all random numbers from the CUDA device:
	float randoms[N_RN];

	// allocate memory on CUDA device:
	float* pDevSeeds;
	cudaMalloc((void**)&pDevSeeds, sizeof(seeds));

	float* pDevRandoms;
	cudaMalloc((void**)&pDevRandoms, sizeof(randoms));

	// copy seeds to CUDA device:
	cudaMemcpy(pDevSeeds, &seeds, sizeof(seeds), cudaMemcpyHostToDevice);

	for(unsigned i=0; i<ITERATIONS; i++)
	{
		// execute kernel function on GPU:
		createRandomNumbers<<<N_BLOCKS, N_THREADS_PER_BLOCK>>>(pDevSeeds, pDevRandoms);

		// copy data back from CUDA Device to 'data' array:
		cudaMemcpy(&randoms, pDevRandoms, sizeof(randoms), cudaMemcpyDeviceToHost);

		// ... do stuff with generated random numbers before next iteration ...
	}

	// free memory on the CUDA Device:
	cudaFree(pDevSeeds);
	cudaFree(pDevRandoms);

	return 0;
}