#include <iostream>
#include <cstdlib>


#define L 256 	// lattice length; must be even number!
#define D 2 	// dimensions
#define N 65536	// number of spins on square lattice (N=L^D)
#define Q 2 	// number of Potts states

#define N_SWEEPS	50000

/* Distribute N/2 spins over threads and blocks:
   since we do "black" and "white" spins seperately, we only need half the threads:
   N_BLOCKS * N_THREADS_PER_BLOCK = N/2 */
#define N_BLOCKS			256  	// number of blocks
#define N_THREADS_PER_BLOCK	128  	// number of threads per block


/**********************************
 *  GPU RANDOM NUMBER GENERATION  *
 **********************************/

__device__ unsigned Tausworthe88(unsigned &z1, unsigned &z2, unsigned &z3)
{
	// Three-step generator with period 2^88
	unsigned b = (((z1 << 13) ^ z1) >> 19);
	z1 = (((z1 & 4294967294) << 12) ^ b);

	b = (((z2 << 2) ^ z2) >> 25);
	z2 = (((z2 & 4294967288) << 4)  ^ b);

	b = (((z3 << 3) ^ z3) >> 11);
	z3 = (((z3 & 4294967280) << 17) ^ b);

	return z1 ^ z2 ^ z3;
}

__device__ unsigned LCRNG(unsigned &z)  
{  
	const unsigned a = 1664525, c = 1013904223;
	return z = a * z + c;
}

__device__ float TauswortheLCRNG(unsigned &z1, unsigned &z2, unsigned &z3, unsigned &z)
{
	// combine both generators and normalize 0...2^32 to 0...1
	return (Tausworthe88(z1, z2, z3) ^ LCRNG(z)) * 2.3283064365e-10;
}


/******************************
 *  GPU METROPOLIS ALGORITHM  *
 ******************************/

__global__ void runMetropolis(int *seeds, unsigned short* spins, unsigned* neighbors, unsigned* spinIdList, int* energyDifferences, float beta)
{
	const unsigned id = blockDim.x*blockIdx.x + threadIdx.x;

	// spin id from list of black or white spins:
	const unsigned spinId = spinIdList[id];

	// get seed values:
	unsigned z1 = seeds[4*id    ];	// Tausworthe seeds
	unsigned z2 = seeds[4*id + 1];
	unsigned z3 = seeds[4*id + 2];
	unsigned z  = seeds[4*id + 3];	// LCRNG seed

	// energy differences for this block:
	__shared__ int deltaE[N_THREADS_PER_BLOCK];
	deltaE[threadIdx.x] = 0;

	unsigned short spinstate = spins[spinId];	// get spin state from DRAM
	unsigned short nb[2*D];						// neighbor states

	// get neighbor states:
	for(unsigned n=0; n<2*D; n++)
		nb[n] = spins[neighbors[2*D*spinId + n]];

	// propose random new spin state:
	unsigned short newstate = floor(TauswortheLCRNG(z1, z2, z3, z) * Q);

	// energy difference: E'-E
	int E_before = 0;
	int E_after  = 0;

	for(unsigned short n=0; n<2*D; n++)
	{
		if(spinstate == nb[n])
			E_before++;
		
		if(newstate == nb[n])
			E_after++;
	}

	// acceptance probability:
	float dE = __int2float_rn(E_before - E_after);
	float pAccept = __expf(-beta*dE);

	if(TauswortheLCRNG(z1, z2, z3, z) <= pAccept)
	{
		spins[spinId] = newstate;   // flip spin
		deltaE[threadIdx.x] = E_before - E_after;	// note energy difference
	}

	// store new seed values in DRAM:
	seeds[4*id    ] = z1;	// Tausworthe seeds
	seeds[4*id + 1] = z2;
	seeds[4*id + 2] = z3;
	seeds[4*id + 3] = z;	// LCRNG seed


	__syncthreads();


	// sum up this block's energy delta:
	if(threadIdx.x == 0)
	{
		int blockEnergyDiff = 0;
		for(unsigned i=0; i<blockDim.x; i++)
			blockEnergyDiff += deltaE[i];

		energyDifferences[blockIdx.x] += blockEnergyDiff;
	}
}



/******************************
 *  HOST FUNCTION (CPU PART)  *
 ******************************/

int main(int argc, char const *argv[])
{
	// each spin has value 0,..,Q-1
	unsigned short spins[N];

	// calculate lattice volume elements:
	unsigned volume[D];
	for(unsigned i=0; i<=D; i++)
	{
		if(i == 0)
			volume[i] = 1;
		else
			volume[i] = volume[i-1] * L;
	}	


	/* Determine the "checkerboard color" (black or white) for each site and
	   initialise lattice with random spin states: */
	unsigned w=0, b=0;
	unsigned white[N/2], black[N/2];	// store ids of white/black sites

	for(unsigned i=0; i<N; i++)
	{
		// Sum of all coordinates even or odd? -> gives checkerboard color
		int csum = 0;
		for(int k=D-1; k>=0; k--)
			csum += ceil((i+1.0)/volume[k]) - 1;

		if((csum%2) == 0)	// white
		{
			white[w] = i;
			w++;
		}
		else				// black
		{
			black[b] = i;
			b++;
		}

		// random spin state:
		spins[i] = floor(drand48() * Q);
	}


	// neighborhood table:
	unsigned neighbors[2*D*N];

	// calculate neighborhood table:
	for(unsigned i=0; i<N; i++)
	{
		unsigned short c=0;

		for(unsigned short dim=0; dim<D; dim++)	// dimension loop
		{
			for(short dir=-1; dir<=1; dir+=2)	// two directions in each dimension
			{
				// neighbor's id in spin list:
				int npos = i + dir * volume[dim];

				// periodic boundary conditions:
				int test = (i % volume[dim+1]) + dir*volume[dim];

				if(test < 0)
					npos += volume[dim+1];
				else if(test >= volume[dim+1])
					npos -= volume[dim+1];
				
				neighbors[2*D*i + c] = npos;
				c++;
			}
		}
	}


	// create 4 seed values for each thread:
	unsigned seeds[4*N/2];
	for(unsigned i=0; i<4*N/2; i++)
	{
		 seeds[i] = static_cast<unsigned>(4294967295 * drand48());
	}


	// calculate energy (Potts model)
	int E = 0;
	for(unsigned i=0; i<N; i++)	
	{
		for(unsigned j=0; j<2*D; j++)
		{
			if(spins[i] == spins[neighbors[2*D*i + j]])
				E--;
		}
	}
	E /= 2; // count each interaction only once


	// copy seeds to GPU:
	int *devPtrSeeds;
	cudaMalloc((void**)&devPtrSeeds, sizeof(seeds));
	cudaMemcpy(devPtrSeeds, &seeds, sizeof(seeds), cudaMemcpyHostToDevice);

	// copy spins to GPU:
	unsigned short *devPtrSpins;
	cudaMalloc((void**)&devPtrSpins, sizeof(spins));
	cudaMemcpy(devPtrSpins, &spins, sizeof(spins), cudaMemcpyHostToDevice);

	// copy neighborhood table to GPU:
	unsigned *devPtrNeighbors;
	cudaMalloc((void**)&devPtrNeighbors, sizeof(neighbors));
	cudaMemcpy(devPtrNeighbors, &neighbors, sizeof(neighbors), cudaMemcpyHostToDevice);

	// copy white ids to GPU:
	unsigned *devPtrWhite;
	cudaMalloc((void**)&devPtrWhite, sizeof(white));
	cudaMemcpy(devPtrWhite, &white, sizeof(white), cudaMemcpyHostToDevice);

	// copy black ids to GPU:
	unsigned *devPtrBlack;
	cudaMalloc((void**)&devPtrBlack, sizeof(black));
	cudaMemcpy(devPtrBlack, &black, sizeof(black), cudaMemcpyHostToDevice);

	// each block calculates energy difference to initial state:
	int energyDifferences[N_BLOCKS];
	for(unsigned i=0; i<N_BLOCKS; i++)
		energyDifferences[i] = 0;

	int *devPtrEnergyDifferences;
	cudaMalloc((void**)&devPtrEnergyDifferences, sizeof(energyDifferences));
	cudaMemcpy(devPtrEnergyDifferences, &energyDifferences, sizeof(energyDifferences), cudaMemcpyHostToDevice);


	std::cout<<"CUDA L="<<L<<"\n";

	int E_before_simulation = E;
	long long sum_E;
	float beta;

	for(float T=0.4; T<=2.63; T+=0.05)
	{
		beta = 1.0f/T;
		sum_E = 0;

		for(unsigned i=0; i<N_SWEEPS; i++)
		{
			// White spins:
			runMetropolis<<<N_BLOCKS, N_THREADS_PER_BLOCK>>>(devPtrSeeds, devPtrSpins, devPtrNeighbors, devPtrWhite, devPtrEnergyDifferences, beta);

			// Black spins:
			runMetropolis<<<N_BLOCKS, N_THREADS_PER_BLOCK>>>(devPtrSeeds, devPtrSpins, devPtrNeighbors, devPtrBlack, devPtrEnergyDifferences, beta);

			// Sum up energy after a thermalization time for mean energy value:
			if(i >= 0.2*N_SWEEPS)
			{
				// get energy changes from the GPU:
				cudaMemcpy(&energyDifferences, devPtrEnergyDifferences, sizeof(energyDifferences), cudaMemcpyDeviceToHost);

				E = E_before_simulation;
				for(unsigned t=0; t<N_BLOCKS; t++)	// take energy changes into account
					E += energyDifferences[t];

				sum_E += E;
			}
		}
	}

	cudaFree(devPtrSeeds);
	cudaFree(devPtrSpins);
	cudaFree(devPtrNeighbors);
	cudaFree(devPtrWhite);
	cudaFree(devPtrBlack);
	cudaFree(devPtrEnergyDifferences);

	return 0;
}